from py_zipkin.zipkin import create_http_headers_for_new_span  # pragma: no cover


# Backwards compatibility for places where pyramid_zipkin is unpinned
create_headers_for_new_span = create_http_headers_for_new_span  # pragma: no cover



import functools
import traceback
import warnings
from collections import namedtuple
from typing import Any
from typing import Callable

from py_zipkin import Encoding
from py_zipkin import Kind
from py_zipkin.exception import ZipkinError
from py_zipkin.storage import get_default_tracer
from py_zipkin.transport import BaseTransportHandler
from pyramid.registry import Registry
from pyramid.request import Request
from pyramid.response import Response

from pyramid_zipkin.request_helper import create_zipkin_attr
from pyramid_zipkin.request_helper import get_binary_annotations
from pyramid_zipkin.request_helper import should_not_sample_path
from pyramid_zipkin.request_helper import should_not_sample_route


def _getattr_path(obj: Any, path: str) -> Any:
    """
    getattr for a dot separated path

    If an AttributeError is raised, it will return None.
    """
    if not path:
        return None

    for attr in path.split('.'):
        obj = getattr(obj, attr, None)
    return obj


_ZipkinSettings = namedtuple('_ZipkinSettings', [
    'zipkin_attrs',
    'transport_handler',
    'service_name',
    'span_name',
    'add_logging_annotation',
    'report_root_timestamp',
    'host',
    'port',
    'context_stack',
    'firehose_handler',
    'post_handler_hook',
    'max_span_batch_size',
    'use_pattern_as_span_name',
    'encoding',
])


def _get_settings_from_request(request: Request) -> _ZipkinSettings:
    """Extracts Zipkin attributes and configuration from request attributes.
    See the `zipkin_span` context in py-zipkin for more detaied information on
    all the settings.

    Here are the supported Pyramid registry settings:

    zipkin.create_zipkin_attr: allows the service to override the creation of
        Zipkin attributes. For example, if you want to deterministically
        calculate trace ID from some service-specific attributes.
    zipkin.transport_handler: how py-zipkin will log the spans it generates.
    zipkin.stream_name: an additional parameter to be used as the first arg
        to the transport_handler function. A good example is a Kafka topic.
    zipkin.add_logging_annotation: if true, the outermost span in this service
        will have an annotation set when py-zipkin begins its logging.
    zipkin.report_root_timestamp: if true, the outermost span in this service
        will set its timestamp and duration attributes. Use this only if this
        service is not going to have a corresponding client span. See
        https://github.com/Yelp/pyramid_zipkin/issues/68
    zipkin.firehose_handler: [EXPERIMENTAL] this enables "firehose tracing",
        which will log 100% of the spans to this handler, regardless of
        sampling decision. This is experimental and may change or be removed
        at any time without warning.
    zipkin.use_pattern_as_span_name: if true, we'll use the pyramid route pattern
        as span name. If false (default) we'll keep using the raw url path.
    """
    settings = request.registry.settings

    # Creates zipkin_attrs and attaches a zipkin_trace_id attr to the request
    if 'zipkin.create_zipkin_attr' in settings:
        zipkin_attrs = settings['zipkin.create_zipkin_attr'](request)
    else:
        zipkin_attrs = create_zipkin_attr(request)

    if 'zipkin.transport_handler' in settings:
        transport_handler = settings['zipkin.transport_handler']
        if not isinstance(transport_handler, BaseTransportHandler):
            warnings.warn(
                'Using a function as transport_handler is deprecated. '
                'Please extend py_zipkin.transport.BaseTransportHandler',
                DeprecationWarning,
            )
            stream_name = settings.get('zipkin.stream_name', 'zipkin')
            transport_handler = functools.partial(transport_handler, stream_name)
    else:
        raise ZipkinError(
            "`zipkin.transport_handler` is a required config property, which"
            " is missing. Have a look at py_zipkin's docs for how to implement"
            " it: https://github.com/Yelp/py_zipkin#transport"
        )

    context_stack = _getattr_path(request, settings.get('zipkin.request_context'))

    service_name = settings.get('service_name', 'unknown')
    span_name = f'{request.method} {request.path}'
    add_logging_annotation = settings.get(
        'zipkin.add_logging_annotation',
        False,
    )

    # If the incoming request doesn't have Zipkin headers, this request is
    # assumed to be the root span of a trace. There's also a configuration
    # override to allow services to write their own logic for reporting
    # timestamp/duration.
    if 'zipkin.report_root_timestamp' in settings:
        report_root_timestamp = settings['zipkin.report_root_timestamp']
    else:
        report_root_timestamp = 'X-B3-TraceId' not in request.headers
    zipkin_host = settings.get('zipkin.host')
    zipkin_port = settings.get('zipkin.port', request.server_port)
    firehose_handler = settings.get('zipkin.firehose_handler')
    post_handler_hook = settings.get('zipkin.post_handler_hook')
    max_span_batch_size = settings.get('zipkin.max_span_batch_size')
    use_pattern_as_span_name = bool(
        settings.get('zipkin.use_pattern_as_span_name', False),
    )
    encoding = settings.get('zipkin.encoding', Encoding.V2_JSON)
    return _ZipkinSettings(
        zipkin_attrs,
        transport_handler,
        service_name,
        span_name,
        add_logging_annotation,
        report_root_timestamp,
        zipkin_host,
        zipkin_port,
        context_stack,
        firehose_handler,
        post_handler_hook,
        max_span_batch_size,
        use_pattern_as_span_name,
        encoding=encoding,
    )


Handler = Callable[[Request], Response]


def zipkin_tween(handler: Handler, registry: Registry) -> Handler:
    """
    Factory for pyramid tween to handle zipkin server logging. Note that even
    if the request isn't sampled, Zipkin attributes are generated and pushed
    into threadlocal storage, so `create_http_headers_for_new_span` and
    `zipkin_span` will have access to the proper Zipkin state.

    Consumes custom create_zipkin_attr function if one is set in the pyramid
    registry.

    :param handler: pyramid request handler
    :param registry: pyramid app registry

    :returns: pyramid tween
    """
    def tween(request: Request) -> Response:
        zipkin_settings = _get_settings_from_request(request)
        tracer = get_default_tracer()

        tween_kwargs = dict(
            service_name=zipkin_settings.service_name,
            span_name=zipkin_settings.span_name,
            zipkin_attrs=zipkin_settings.zipkin_attrs,
            transport_handler=zipkin_settings.transport_handler,
            host=zipkin_settings.host,
            port=zipkin_settings.port,
            add_logging_annotation=zipkin_settings.add_logging_annotation,
            report_root_timestamp=zipkin_settings.report_root_timestamp,
            context_stack=zipkin_settings.context_stack,
            max_span_batch_size=zipkin_settings.max_span_batch_size,
            encoding=zipkin_settings.encoding,
            kind=Kind.SERVER,
        )

        # Only set the firehose_handler if it's defined and only if the current
        # request is not blacklisted. This prevents py_zipkin from emitting
        # firehose spans for blacklisted paths like /status
        if zipkin_settings.firehose_handler is not None and \
                not should_not_sample_path(request) and \
                not should_not_sample_route(request):
            tween_kwargs['firehose_handler'] = zipkin_settings.firehose_handler

        with tracer.zipkin_span(**tween_kwargs) as zipkin_context:
            response = None
            try:
                response = handler(request)
            except Exception as e:
                exception_type = type(e).__name__
                exception_stacktrace = traceback.format_exc()
                zipkin_context.update_binary_annotations({
                    'error.type': exception_type,
                    'response_status_code': '500',
                    'http.response.status_code': '500',
                    'exception.stacktrace': exception_stacktrace,
                })
                zipkin_context.add_annotation(exception_type)
                raise e
            finally:
                if zipkin_settings.use_pattern_as_span_name \
                        and request.matched_route:
                    zipkin_context.override_span_name('{} {}'.format(
                        request.method,
                        request.matched_route.pattern,
                    ))
                zipkin_context.update_binary_annotations(
                    get_binary_annotations(request, response),
                )

                if zipkin_settings.post_handler_hook:
                    zipkin_settings.post_handler_hook(
                        request,
                        response,
                        zipkin_context
                    )

            return response

    return tween



import random
import re
from typing import Dict
from typing import Optional

from py_zipkin.util import generate_random_128bit_string
from py_zipkin.util import generate_random_64bit_string
from py_zipkin.zipkin import ZipkinAttrs
from pyramid.interfaces import IRoutesMapper
from pyramid.request import Request
from pyramid.response import Response

from pyramid_zipkin.version import __version__


DEFAULT_REQUEST_TRACING_PERCENT = 0.5


def get_trace_id(request: Request) -> str:
    """Gets the trace id based on a request. If not present with the request, a
    completely random 128-bit trace id is generated.

    :param: current active pyramid request
    :returns: the value of the 'X-B3-TraceId' header or a 128-bit hex string
    """
    return request.headers.get('X-B3-TraceId', generate_random_128bit_string())


def should_not_sample_path(request: Request) -> bool:
    """Decided whether current request path should be sampled or not. This is
    checked previous to `should_not_sample_route` and takes precedence.

    :param: current active pyramid request
    :returns: boolean whether current request path is blacklisted.
    """
    blacklisted_paths = request.registry.settings.get(
        'zipkin.blacklisted_paths', [])
    # Only compile strings, since even recompiling existing
    # compiled regexes takes time.
    regexes = [
        re.compile(r) if isinstance(r, str) else r
        for r in blacklisted_paths
    ]
    return any(r.match(request.path) for r in regexes)


def should_not_sample_route(request: Request) -> bool:
    """Decided whether current request route should be sampled or not.

    :param: current active pyramid request
    :returns: boolean whether current request route is blacklisted.
    """
    blacklisted_routes = request.registry.settings.get(
        'zipkin.blacklisted_routes', [])

    if not blacklisted_routes:
        return False
    route_mapper = request.registry.queryUtility(IRoutesMapper)
    route_info = route_mapper(request).get('route')
    return (route_info and route_info.name in blacklisted_routes)


def should_sample_as_per_zipkin_tracing_percent(tracing_percent: float) -> bool:
    """Calculate whether the request should be traced as per tracing percent.

    :param tracing_percent: value between 0.0 to 100.0
    :type tracing_percent: float
    :returns: boolean whether current request should be sampled.
    """
    return (random.random() * 100) < tracing_percent


def is_tracing(request: Request) -> bool:
    """Determine if zipkin should be tracing
    1) Check whether the current request path is blacklisted.
    2) If not, check whether the current request route is blacklisted.
    3) If not, check if specific sampled header is present in the request.
    4) If not, Use a tracing percent (default: 0.5%) to decide.

    :param request: pyramid request object

    :returns: boolean True if zipkin should be tracing
    """
    if should_not_sample_path(request):
        return False
    elif should_not_sample_route(request):
        return False
    elif 'X-B3-Sampled' in request.headers:
        return request.headers.get('X-B3-Sampled') == '1'
    else:
        zipkin_tracing_percent = request.registry.settings.get(
            'zipkin.tracing_percent', DEFAULT_REQUEST_TRACING_PERCENT)
        return should_sample_as_per_zipkin_tracing_percent(
            zipkin_tracing_percent)


def create_zipkin_attr(request: Request) -> ZipkinAttrs:
    """Create ZipkinAttrs object from a request with sampled flag as True.
    Attaches lazy attribute `zipkin_trace_id` with request which is then used
    throughout the tween.

    Consumes custom is_tracing function to determine if the request is traced
    if one is set in the pyramid registry.

    :param request: pyramid request object
    :rtype: :class:`pyramid_zipkin.request_helper.ZipkinAttrs`
    """
    settings = request.registry.settings

    if 'zipkin.is_tracing' in settings:
        is_sampled = settings['zipkin.is_tracing'](request)
    else:
        is_sampled = is_tracing(request)

    span_id = request.headers.get(
        'X-B3-SpanId', generate_random_64bit_string())
    parent_span_id = request.headers.get('X-B3-ParentSpanId', None)
    flags = request.headers.get('X-B3-Flags', '0')

    # Store zipkin_trace_id and zipkin_span_id in the request object so that
    # they're still available once we leave the pyramid_zipkin tween. An example
    # is being able to log them in the pyramid exc_logger, which runs after all
    # tweens have been exited.
    request.zipkin_trace_id = get_trace_id(request)
    request.zipkin_span_id = span_id

    return ZipkinAttrs(
        trace_id=request.zipkin_trace_id,
        span_id=span_id,
        parent_span_id=parent_span_id,
        flags=flags,
        is_sampled=is_sampled,
    )


def get_binary_annotations(
    request: Request,
    response: Response,
) -> Dict[str, Optional[str]]:
    """Helper method for getting all binary annotations from the request.

    :param request: the Pyramid request object
    :param response: the Pyramid response object
    :returns: binary annotation dict of {str: str}
    """
    # use only @property of the request object
    # https://sourcegraph.com/search?q=repo:%5Egithub%5C.com/Pylons/webob$@1.8.7++file:%5Esrc/webob/request%5C.py$+@property&patternType=keyword&sm=0
    annotations = {
        'http.uri': request.path,
        'http.uri.qs': request.path_qs,
        'otel.library.name': __name__.split('.')[0],
        'otel.library.version': __version__,
    }

    if request.client_addr:
        annotations['client.address'] = request.client_addr

    if request.matched_route:
        annotations['http.route'] = request.matched_route.pattern

    # update attributes from request.environ object
    # https://sourcegraph.com/github.com/Pylons/webob@1.8.7/-/blob/docs/index.txt?L63-68
    request_env = request.environ
    _update_annotations_from_request_environ(request_env, annotations)

    if response:
        status_code = response.status_code
        if isinstance(status_code, int):
            annotations['http.response.status_code'] = str(status_code)
            annotations['response_status_code'] = str(status_code)
            if 100 <= status_code < 200:
                annotations['otel.status_code'] = 'Unset'
            elif 200 <= status_code < 300:
                annotations['otel.status_code'] = 'Ok'
            elif 300 <= status_code < 500:
                annotations['otel.status_code'] = 'Unset'
            else:
                annotations['otel.status_code'] = 'Error'

        else:
            annotations['otel.status_code'] = 'Error'
            annotations['otel.status_description'] = (
                f'Non-integer HTTP status code: {repr(status_code)}'
            )

    settings = request.registry.settings
    if 'zipkin.set_extra_binary_annotations' in settings:
        annotations.update(
            settings['zipkin.set_extra_binary_annotations'](request, response)
        )
    return annotations


def _update_annotations_from_request_environ(
        environ: Dict,
        annotations: Dict[str, Optional[str]]
) -> None:
    method = environ.get('REQUEST_METHOD', '').strip()
    if method:
        annotations['http.request.method'] = method

    flavor = environ.get('SERVER_PROTOCOL', '')
    if flavor:
        annotations['network.protocol.version'] = flavor

    path = environ.get('PATH_INFO')
    if path:
        annotations['url.path'] = path

    host_name = environ.get('SERVER_NAME')
    host_port = environ.get('SERVER_PORT')

    if host_name:
        annotations['server.address'] = host_name

    if host_port:
        annotations['server.port'] = str(host_port)

    url_scheme = environ.get('wsgi.url_scheme')
    if url_scheme:
        annotations['url.scheme'] = url_scheme

    user_agent = environ.get('HTTP_USER_AGENT')
    if user_agent:
        annotations['user_agent.original'] = user_agent

    query_string = environ.get('QUERY_STRING')
    if query_string:
        annotations['url.query'] = query_string



from pyramid.config import Configurator
from pyramid.tweens import INGRESS


def includeme(config: Configurator) -> None:  # pragma: no cover
    """
    :type config: :class:`pyramid.config.Configurator`
    """
    config.add_tween('pyramid_zipkin.tween.zipkin_tween', under=INGRESS)



#!/usr/bin/python
from setuptools import find_packages
from setuptools import setup

with open('pyramid_zipkin/version.py') as f:
    exec(f.read())

version = locals()['__version__']

setup(
    name='pyramid_zipkin',
    version=version,
    provides=["pyramid_zipkin"],
    author='Yelp, Inc.',
    author_email='opensource+pyramid-zipkin@yelp.com',
    license='Copyright Yelp 2018',
    url="https://github.com/Yelp/pyramid_zipkin",
    description='Zipkin instrumentation for the Pyramid framework.',
    packages=find_packages(exclude=('tests*',)),
    package_data={
        '': ['*.thrift'],
        'pyramid_zipkin': ['py.typed'],
    },
    install_requires=[
        'py_zipkin >= 0.18.1',
        'pyramid',
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.8",
)
